"""Utility functions for executing Schedules on Qblox hardware."""

from __future__ import annotations

from typing import TYPE_CHECKING

from qcodes.instrument import find_or_create_instrument

from qblox_instruments import Cluster, ClusterType
from quantify_core.measurement.control import MeasurementControl
from quantify_core.visualization.instrument_monitor import InstrumentMonitor
from quantify_core.visualization.pyqt_plotmon import PlotMonitor_pyqt as PlotMonitor
from quantify_scheduler import GenericInstrumentCoordinatorComponent, MockLocalOscillator
from quantify_scheduler.backends.graph_compilation import SerialCompiler
from quantify_scheduler.backends.types.qblox import ClusterDescription
from quantify_scheduler.instrument_coordinator import InstrumentCoordinator
from quantify_scheduler.instrument_coordinator.components.qblox import ClusterComponent
from quantify_scheduler.schedules.schedule import CompiledSchedule, Schedule

if TYPE_CHECKING:
    from xarray import Dataset

    from quantify_scheduler import QuantumDevice
    from quantify_scheduler.backends.qblox_backend import QbloxHardwareCompilationConfig


STR_TO_DUMMY_TYPE = {
    "QCM": ClusterType.CLUSTER_QCM,
    "QRM": ClusterType.CLUSTER_QRM,
    "QCM_RF": ClusterType.CLUSTER_QCM_RF,
    "QRM_RF": ClusterType.CLUSTER_QRM_RF,
    "QTM": ClusterType.CLUSTER_QTM,
    "QDM": ClusterType.CLUSTER_QDM,
}


def initialize_hardware(
    quantum_device: QuantumDevice | None = None,
    ip: str | None = None,
    live_plotting: bool = False,
) -> tuple[MeasurementControl, InstrumentCoordinator, Cluster]:
    """
    Initialize MeasurementControl and InstrumentCoordinator from QuantumDevice.

    Including Cluster instruments and Mock Local Oscillators for specific non-cluster devices.

    Parameters
    ----------
    quantum_device : QuantumDevice | None, optional
        target quantum device, by default None
    ip : str | None, optional
        ip address of the qblox cluster. Will use a dummy cluster if None, by default None
    live_plotting : bool, optional
        whether live plotting should be enabled, by default False

    Returns
    -------
    tuple[MeasurementControl, InstrumentCoordinator, Cluster]

    Raises
    ------
    ValueError
        Neither QuantumDevice nor global default are provided.

    """
    config: QbloxHardwareCompilationConfig = quantum_device.generate_hardware_compilation_config()  # type: ignore
    all_instruments = []
    for name, description in config.hardware_description.items():
        if isinstance(description, ClusterDescription):
            cluster = _cluster_from_cluster_description(ip, name, description)
            all_instruments.append(cluster)
        # Check for other instrument types and initialize as MockLocalOscillator if needed (crucial for OAQ)
        elif description.instrument_type in ["OpticalModulator", "LocalOscillator"]:
            # Dynamically create a variable with the instrument's name as the variable name
            globals()[name] = MockLocalOscillator(name=name)
            all_instruments.append(globals()[name])

    meas_ctrl = find_or_create_instrument(MeasurementControl, recreate=True, name="meas_ctrl")
    quantum_device.instr_measurement_control(meas_ctrl.name)
    ic = find_or_create_instrument(InstrumentCoordinator, recreate=True, name="ic")
    quantum_device.instr_instrument_coordinator(ic.name)

    for instrument in all_instruments:
        # Add each instrument to the instrument coordinator
        if isinstance(instrument, Cluster):
            ic_cluster = ClusterComponent(instrument)
            ic.add_component(ic_cluster)
        else:  # For MockLocalOscillators
            ic_component = GenericInstrumentCoordinatorComponent(instrument)
            ic.add_component(ic_component)

    if live_plotting:
        pm = find_or_create_instrument(PlotMonitor, recreate=True, name="PlotMonitor")
        meas_ctrl.instr_plotmon(pm.name)
        find_or_create_instrument(InstrumentMonitor, recreate=True, name="Instrument_Monitor")

    return (meas_ctrl, ic, cluster)


def _cluster_from_cluster_description(
    ip: str | None, name: str, description: ClusterDescription
) -> Cluster:
    instrument_ip = ip if ip else getattr(description, "ip", None)

    if instrument_ip:
        dummy_config = None
    else:
        dummy_config = {
            int(slot): STR_TO_DUMMY_TYPE[module.instrument_type]
            for slot, module in description.modules.items()
        }
    cluster = find_or_create_instrument(
        Cluster,
        recreate=True,
        name=name,
        identifier=instrument_ip,
        dummy_cfg=dummy_config,
    )

    return cluster


def run_schedule(schedule: Schedule, quantum_device: QuantumDevice) -> Dataset:
    if not isinstance(schedule, CompiledSchedule):
        compiler = SerialCompiler(name="auto_compiler", quantum_device=quantum_device)
        schedule = compiler.compile(schedule=schedule)
    ic: InstrumentCoordinator = quantum_device.instr_instrument_coordinator.get_instr()
    ic.prepare(schedule)
    ic.start()
    return ic.retrieve_acquisition()
